import struct
from typing import BinaryIO

from nwn._shared import get_nwn_encoding, GenderedLanguage
from nwn.gff._types import (
    Dword,
    CExoString,
    ResRef,
    CExoLocString,
    Struct,
    List,
    Double,
    Int64,
    Dword64,
    VOID,
    SIMPLE_TYPES,
)
from nwn.gff._impl import FieldKind, Header, FieldEntry, StructEntry


def read(file: BinaryIO):
    """
    Read a GFF data from a binary stream.

    Example:
        >>> with open("file.gff", "rb") as f:
        ...     root, file_type = gff.read(f)
        ...     print(root)

    Args:
        file: The binary stream to read from.

    Returns:
        A tuple containing the root struct and the file type.
    """

    root_offset = file.tell()
    labels = []
    fields = []
    structs = []
    list_indices = []
    field_indices = []
    resolved_structs = {}
    struct_parents = {}

    header = Header(
        file.read(4).decode("ascii"),
        file.read(4).decode("ascii"),
        *struct.unpack("<12i", file.read(48)),
    )
    if header.file_version != "V3.2":
        raise ValueError(f"Unsupported GFF version: {header.file_version}")

    file.seek(root_offset + header.label_offset)
    for _ in range(header.label_count):
        labels.append(file.read(16).split(b"\x00")[0].decode("ascii"))

    file.seek(root_offset + header.field_offset)
    for _ in range(header.field_count):
        data = struct.unpack("<III", file.read(12))
        fields.append(FieldEntry(FieldKind(data[0]), data[1], data[2]))

    file.seek(root_offset + header.field_indices_offset)
    for _ in range(header.field_indices_size // 4):
        field_indices.append(struct.unpack("<I", file.read(4))[0])

    file.seek(root_offset + header.list_indices_offset)
    for _ in range(header.list_indices_size // 4):
        list_indices.append(struct.unpack("<I", file.read(4))[0])

    file.seek(root_offset + header.struct_offset)
    for _ in range(header.struct_count):
        data = struct.unpack("<III", file.read(12))
        structs.append(StructEntry(data[0], data[1], data[2]))

    def _read_field_value(field):
        if field.type in SIMPLE_TYPES:
            cls = SIMPLE_TYPES[field.type]
            us = cls.SIMPLE_DATA_FORMAT
            data = struct.pack("<I", field.data_or_offset)[: struct.calcsize(us)]
            up = struct.unpack("<" + us, data)
            return cls(up[0])

        file.seek(root_offset + header.field_data_offset + field.data_or_offset)

        if field.type == FieldKind.DOUBLE:
            return Double(struct.unpack("<d", file.read(8))[0])

        if field.type == FieldKind.DWORD64:
            return Dword64(struct.unpack("<Q", file.read(8))[0])

        if field.type == FieldKind.INT64:
            return Int64(struct.unpack("<q", file.read(8))[0])

        if field.type == FieldKind.CEXOSTRING:
            sz = struct.unpack("<I", file.read(4))[0]
            if sz > 0xFFFF:
                raise ValueError("String too long")
            return CExoString(file.read(sz).decode(get_nwn_encoding()))

        if field.type == FieldKind.RESREF:
            sz = struct.unpack("<b", file.read(1))[0]
            if sz > 16:
                raise ValueError("Resref too long")
            return ResRef(file.read(sz).decode(get_nwn_encoding()))

        if field.type == FieldKind.CEXOLOCSTRING:
            _ = struct.unpack("<I", file.read(4))[0]
            strref = Dword(struct.unpack("<I", file.read(4))[0])
            count = struct.unpack("<I", file.read(4))[0]
            entries = {}
            for _ in range(count):
                fid = GenderedLanguage.from_id(struct.unpack("<I", file.read(4))[0])
                sz = struct.unpack("<I", file.read(4))[0]
                entries[fid] = file.read(sz).decode(get_nwn_encoding())
            return CExoLocString(strref, entries)

        if field.type == FieldKind.VOID:
            sz = struct.unpack("<I", file.read(4))[0]
            return VOID(file.read(sz))

        if field.type == FieldKind.LIST:
            offset = field.data_or_offset // 4
            size = list_indices[offset]
            start = offset + 1
            end = start + size

            return List([_read_struct(field, lid) for lid in list_indices[start:end]])

        if field.type == FieldKind.STRUCT:
            return _read_struct(field, field.data_or_offset)

        raise NotImplementedError(f"Field {field} not implemented")

    def _read_struct(parent, struct_idx) -> Struct:
        if struct_idx in resolved_structs:
            if struct_parents[struct_idx] != parent:
                raise ValueError("Struct already resolved with different parent")
            return resolved_structs[struct_idx]

        struct_entry = structs[struct_idx]

        if struct_entry.field_count == 1:
            field_array_indices = [struct_entry.data_or_offset]
        else:
            start = struct_entry.data_or_offset // 4
            end = start + struct_entry.field_count
            if end < start:
                raise ValueError("Field index array out of bounds")

            field_array_indices = field_indices[start:end]

        resolved_structs[struct_idx] = Struct(
            struct_entry.id,
            **{
                labels[fld.label_index]: _read_field_value(fld)
                for fld in map(lambda x: fields[x], field_array_indices)
            },
        )

        struct_parents[struct_idx] = parent
        return resolved_structs[struct_idx]

    root = _read_struct(None, 0)
    return root, header.file_type



"""
Read and write TLK (Talk Table) files (for translation and base game string references).

All strings are transparently converted to and from the NWN encoding.
"""

import struct

from typing import NamedTuple, BinaryIO
from nwn._shared import Language, get_nwn_encoding


class TlkEntry(NamedTuple):
    """
    A single entry in a TLK file including the text, sound resref, and sound length.
    Only used when reading TLK files with sound data.
    """

    text: str
    sound_resref: str = ""
    sound_length: float = 0.0

    def __str__(self):
        return self.text


def read(
    file: BinaryIO, include_sound_data=False, max_entries=0x7FFFF
) -> tuple[list, Language]:
    """
    Reads a TLK file fully into memory and returns a list of entries.

    Args:
        file: A binary file object containing the TLK file.
        include_sound_data: If True, entries are returned as TlkEntry objects.
            Otherwise, the resulting dict will contain only the text.
        max_entries: The maximum number of entries to read from the TLK file.
            This is a sanity check to avoid allocating excess memory when
            reading untrusted or corrupted data.

    Returns:
        A tuple containing a list of entries and the language of the TLK file.
            The list contains either strings or TlkEntry objects, depending on
            the value of include_sound_data.

    Raises:
        ValueError: If the file does not contain valid TLK data.
    """

    magic = file.read(4)
    if magic != b"TLK ":
        raise ValueError("Invalid TLK magic")
    version = file.read(4)
    if version != b"V3.0":
        raise ValueError("Invalid TLK version")
    language = struct.unpack("<I", file.read(4))[0]
    language = Language(language)
    entry_count = struct.unpack("<I", file.read(4))[0]
    entries_offset = struct.unpack("<I", file.read(4))[0]

    if entry_count > max_entries:
        raise ValueError(f"Too many entries in TLK file: {entry_count} > {max_entries}")

    entries = []
    for i in range(entry_count):
        file.seek(20 + 40 * i)

        (
            _,  # flags (text_present=0x1, sound_present=0x2,length_present=0x4)
            sound_resref,
            _,  # volume variance: unused as per spec
            _,  # pitch variance: unused as per spec
            offset_to_string,
            string_sz,
            sound_length,
        ) = struct.unpack("<I16sIIIIf", file.read(40))

        sound_resref = sound_resref.decode("ascii").strip("\x00\xc0")

        file.seek(entries_offset + offset_to_string)
        text = file.read(string_sz).decode(get_nwn_encoding())

        if include_sound_data:
            entries.append(TlkEntry(text, sound_resref, sound_length))
        else:
            entries.append(text)

    return (entries, language)


def write(file: BinaryIO, entries: list, language: Language):
    """
    Writes a Tlk object to a binary file.

    Args:
        file: A binary file object to write the TLK data to.
        entries: A list containing the entries to write, in order.
            Entries can be either strings or TlkEntry objects.
        language: The language of the TLK file to write.

    Raises:
        ValueError: If the TLK object contains invalid data.
    """

    file.write(
        struct.pack(
            "<4s4sIII",
            b"TLK ",
            b"V3.0",
            language.value,
            len(entries),
            20 + len(entries) * 40,
        )
    )

    str_data = bytearray()

    string_offset = 0
    for idx, entry in enumerate(entries):
        flags = 0
        text_len = 0

        if isinstance(entry, str):
            entry = TlkEntry(entry)

        if entry.text:
            flags |= 0x1
            text_len = len(entry.text.encode(get_nwn_encoding()))
        if entry.sound_resref:
            flags |= 0x2
        if entry.sound_length:
            flags |= 0x4

        if len(entry.sound_resref) > 16:
            raise ValueError(f"Sound resref at {idx} is too long")

        file.write(
            struct.pack(
                "<I16sIIIIf",
                flags,
                entry.sound_resref.ljust(16, "\x00").encode("ascii"),
                0,  # volume variance: unused as per spec
                0,  # pitch variance: unused as per spec
                string_offset,
                text_len,
                entry.sound_length,
            )
        )
        string_offset += text_len
        str_data += entry.text.encode(get_nwn_encoding())

    file.write(str_data)



"""
A simplified parser for nwscript.nss.
"""

import re
from typing import TextIO, NamedTuple, Any

from ._types import Vector, Object, Location, VMType


class Constant(NamedTuple):
    """
    A constant definition from the language spec (e.g. ``const int TRUE = 1;``).
    """

    ty: VMType
    name: str
    value: Any


class FunctionArg(NamedTuple):
    ty: VMType
    name: str
    default: Any | None


class Function(NamedTuple):
    id: int
    return_type: VMType
    name: str
    args: list[FunctionArg]
    doc: str = ""


class LanguageSpec(NamedTuple):
    constants: list[Constant]
    functions: list[Function]


_VM_TYPE_RX = "(?:" + ("|".join([f.value for f in VMType])) + ")"

_CONST_RX = re.compile("(" + _VM_TYPE_RX + ")" + r"\s+([A-Z0-9_]+)\s*=\s*([^;]+);")

_FUNC_RX = re.compile(r"\b(" + _VM_TYPE_RX + r")\s+(\w+)\s*\((.*?)\)\s*;")

_FUNC_ARG_RX = re.compile(
    r"\s*("
    + _VM_TYPE_RX
    + r")\s+(\w+)(?:\s*=\s*("
    + re.escape(r"[0.0,0.0,0.0]")
    + r"|[\w\"-]+))?"
)


def read(file: TextIO) -> LanguageSpec:
    """
    Reads a language specification from a given file.

    Args:
        file: A file-like object containing the language specification; usually nwscript.nss.

    Returns:
        LanguageSpec: An object containing constants and functions parsed from the file.

    Raises:
        ValueError: If the parsing fails.

    """

    consts = []
    funcs = []

    curdoc = []

    func_idx = 0
    for line in file:
        if line.startswith("//"):
            curdoc.append(line.removeprefix("//").strip())
            continue

        if m := _CONST_RX.match(line):
            ty = VMType(m.group(1))
            va = m.group(3)
            if ty == VMType.FLOAT:
                va = va.strip("f")
            if ty == VMType.INT and va == "TRUE":
                va = 1
            if ty == VMType.INT and va == "FALSE":
                va = 0
            va = ty.python_type()(va)
            consts.append(Constant(ty, m.group(2), va))
            curdoc = []

        if m := _FUNC_RX.match(line):
            args = []
            mm = _FUNC_ARG_RX.findall(m.group(3))
            if len(m.group(3)) != 0 and not mm:
                raise ValueError(f"Failed to parse {m.group(3)}")

            for mmm in mm:
                default = None
                if mmm[2] == "":
                    pass
                elif mmm[2] == "TRUE":
                    default = True
                elif mmm[2] == "FALSE":
                    default = False
                elif mmm[2] == "OBJECT_SELF":
                    default = Object.SELF
                elif mmm[2] == "OBJECT_INVALID":
                    default = Object.INVALID
                elif re.match(r"^-?\d+$", mmm[2]):
                    default = int(mmm[2])
                elif re.match(r"^-?\d+\.\d+f?$", mmm[2]):
                    default = float(mmm[2])
                elif mmm[2] == "[0.0,0.0,0.0]":
                    default = Vector(0, 0, 0)
                elif re.match(r'^"([^"]*)"$', mmm[2]):
                    default = mmm[2][1:-1]
                elif mmm[2] == "LOCATION_INVALID":
                    default = Location.INVALID
                elif re.match(r"^[A-Z][A-Z_]+[A-Z]$", mmm[2]):
                    # At this point, all constants must have been defined
                    c = next((c for c in consts if c.name == mmm[2]), None)
                    if c is None:
                        raise ValueError(
                            f"Unknown constant {mmm[2]} referenced in {m.group(2)}"
                        )
                    default = c.value
                else:
                    raise NotImplementedError(f"Unknown default value {mmm[2]}")
                args.append(FunctionArg(VMType(mmm[0]), mmm[1], default))

            funcs.append(
                Function(func_idx, VMType(m.group(1)), m.group(2), args, curdoc)
            )
            curdoc = []
            func_idx += 1

    return LanguageSpec(consts, funcs)



"""
Read and write 2DA files (2-dimensional array, similar to CSV).
"""

from typing import TextIO

_MAGIC = "2DA V2.0"

CELL = str | None
r"""A type alias for a cell value, which can be a string or None (\*\*\*\*)."""


def _escape_cell(cell: CELL):
    if cell is None:
        return "****"
    if " " in cell:
        return f'"{cell}"'
    return cell


def _split_twoda_style(line: str) -> list[CELL]:
    in_quotes = False
    current = []
    result = []
    for char in line:
        if char == '"':
            in_quotes = not in_quotes
        elif char.isspace() and not in_quotes:
            if current:
                result.append("".join(current))
                current = []
        else:
            current.append(char)
    if current:
        result.append("".join(current))
    return result


class DictReader:
    def __init__(self, columns, f):
        self._columns = columns
        self._f = f

    @property
    def columns(self):
        return self._columns

    def __iter__(self):
        return self

    def __next__(self):
        ln = self._f.readline()
        if not ln:
            raise StopIteration
        ln = _split_twoda_style(ln.strip())
        # First cell is the row number, which canonically gets discarded
        ln.pop(0)
        ln = [None if x == "****" else x for x in ln]
        # Fill in None values for missing columns
        ln = ln + [None] * (len(self._columns) - len(ln))
        return dict(zip(self._columns, ln))

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self._f.close()


def read(file: TextIO):
    """
    Reads a 2DA file and returns a DictReader object.

    Example:
        >>> with open("file.2da", "r") as f:
        ...     for row in twoda.read(f):
        ...         print(row)

    Args:
        file: The 2DA file to read.

    Returns:
        DictReader: An object that allows reading the 2DA file as a dictionary.

    Raises:
        ValueError: Parse/format errors.
    """

    def fetch():
        ln = file.readline()
        if ln is None:
            raise ValueError("Unexpected EOF")
        return ln.strip()

    head = fetch()
    if head != _MAGIC:
        raise ValueError("Not a 2DA file")

    if fetch() != "":
        raise ValueError("Expected empty line after magic")

    columns = fetch().split()

    if not columns:
        raise ValueError("No columns found")

    return DictReader(columns=columns, f=file)


class DictWriter:
    def __init__(self, columns, f):
        self._columns = columns
        self._f = f
        self._idx = 0
        f.write(_MAGIC + "\n\n")
        f.write(" " + " ".join(columns) + "\n")

    def add_row(self, row: dict[str, CELL]):
        self._f.write(
            str(self._idx)
            + " "
            + " ".join(_escape_cell(row.get(h)) for h in self._columns)
            + "\n"
        )
        self._idx += 1


def write(file: TextIO, columns: list[str]) -> DictWriter:
    """
    Writes data to a file using the specified columns.

    Example:
        >>> with open("file.2da", "w") as f:
        ...     writer = twoda.write(f, ["col1", "col2"])
        ...     writer.add_row({"col1": "foo", "col2": "bar"})
        ...     writer.add_row({"col1": "baz", "col2": "qux"})

    Args:
        file: The file object where the data will be written.
        columns: A list of column names to be used in the output file.

    Returns:
        DictWriter: An instance of DictWriter initialized with the given columns and file.
    """

    return DictWriter(columns=columns, f=file)



"""
nwn
===

A python package with various modules for accessing Neverwinter Nights: Enhanced Edition
data formats and functionality.

Stability
---------

This package is currently in ALPHA state. API stability is not guaranteed.

Installation
------------

The package is available on PyPI and can be installed with pip:

.. code-block:: bash

    pip install nwn

License
-------

This package is licensed under the MIT license.
"""

from ._shared import (
    Language,
    GenderedLanguage,
    Gender,
    get_nwn_encoding,
    restype_to_extension,
    extension_to_restype,
    FileMagic,
)

__all__ = [
    "Language",
    "GenderedLanguage",
    "Gender",
    "get_nwn_encoding",
    "restype_to_extension",
    "extension_to_restype",
    "FileMagic",
]



"""Read and write ERF (Encapsulated Resource Format) archives."""

import struct
from typing import NamedTuple, BinaryIO
from enum import Enum
from datetime import date, timedelta

from ._shared import (
    get_nwn_encoding,
    GenderedLanguage,
    restype_to_extension,
    extension_to_restype,
    FileMagic,
)


class Reader:
    """
    Class to read and access an ERF archive (MOD, HAK, ERF, ...)

    Example:
        >>> with open("Prelude.mod", "rb") as file:
        ...    erf = Reader(file)
        ...    print(erf.filenames)
        ...    gff_data = erf.read_file("item.uti")
        ...    ...

    Args:
        file: The file object to read from.
    """

    class Version(Enum):
        V_1_0 = "V1.0"
        # E_1_0 = "E1.0"

    class _Header(NamedTuple):
        file_type: FileMagic
        file_version: "Reader.Version"
        locstr_count: int
        locstr_sz: int
        entry_count: int
        offset_to_locstr: int
        offset_to_keylist: int
        offset_to_reslist: int
        build_year: int
        build_day: int
        description_strref: int

    class Entry(NamedTuple):
        resref: str
        restype: int
        offset: int
        disk_size: int
        uncompressed_size: int

        @property
        def filename(self):
            return f"{self.resref.lower()}.{restype_to_extension(self.restype)}"

    def _seek(self, relative_to_start):
        self._file.seek(self._root_offset + relative_to_start)

    def __init__(self, file, max_entries=65535, max_locstr=100):
        self._file = file
        self._root_offset = self._file.tell()

        ft = self._file.read(4)
        fv = self.Version(self._file.read(4).decode("ASCII"))
        va = struct.unpack("IIIIIIIII", self._file.read(36))
        self._header = self._Header(FileMagic(ft), fv, *va)

        if self._header.entry_count > max_entries:
            raise ValueError("Too many resources")

        if self._header.locstr_count > max_locstr:
            raise ValueError("Too many localized strings")

        self._seek(self._header.offset_to_locstr)
        loc_str = {}
        for _ in range(self._header.locstr_count):
            lid = GenderedLanguage.from_id(struct.unpack("I", self._file.read(4))[0])
            sz = struct.unpack("I", self._file.read(4))[0]
            st = self._file.read(sz).decode(get_nwn_encoding())
            loc_str[lid] = st

        self._seek(self._header.offset_to_reslist)
        resources = []
        for _ in range(self._header.entry_count):
            offset = struct.unpack("i", self._file.read(4))[0]
            disk_size = struct.unpack("i", self._file.read(4))[0]
            uncompressed = disk_size
            resources.append((offset, disk_size, uncompressed))

        self._seek(self._header.offset_to_keylist)
        keys = []
        for _ in range(self._header.entry_count):
            resref = self._file.read(16).split(b"\x00")[0].decode("ASCII")
            _ = struct.unpack("I", self._file.read(4))[0]  # res_id unused
            res_type = struct.unpack("H", self._file.read(2))[0]
            _ = self._file.read(2)  # unused
            keys.append((resref, res_type))

        self._localized_strings = loc_str

        self._files = {
            f"{resref.lower()}.{restype_to_extension(restype)}": self.Entry(
                resref, restype, o, d, u
            )
            for (resref, restype), (o, d, u) in zip(keys, resources)
        }

    @property
    def file_type(self) -> FileMagic:
        """The file type of the ERF archive."""
        return self._header.file_type

    @property
    def build_date(self) -> date:
        """The build date of the ERF archive."""
        return date(1900 + self._header.build_year, 1, 1) + timedelta(
            days=self._header.build_day
        )

    @property
    def localized_strings(self) -> dict[GenderedLanguage, str]:
        """The localized strings in the ERF archive."""
        return self._localized_strings

    @property
    def description_strref(self) -> int:
        """The STRREF set in the ERF header. 0 if not set."""
        return self._header.description_strref

    @property
    def filenames(self) -> list[str]:
        """
        Returns the filenames in the ERF archive.

        Returns:
            A list of filenames present in the ERF archive.
        """
        return list(self._files.keys())

    @property
    def filemap(self) -> dict[str, Entry]:
        """
        Returns the mapping of files.

        This method returns the internal dictionary that maps file names to their
        corresponding file data.

        Returns:
            A dict mapping file names to their corresponding Entry objects.
        """
        return self._files

    def read_file(self, filename: str) -> bytes:
        """
        Retrieve the contents of a file from the archive.

        Args:
            filename: The name of the file to retrieve.

        Returns:
            The contents of the file as a byte string.

        Raises:
            KeyError: If the file is not found in the archive.
            ValueError: If the filename is of a unknown restype.
        """
        resource = self._files[filename.lower()]
        self._seek(resource.offset)
        return self._file.read(resource.disk_size)


class Writer:
    """
    A class to write ERF files.

    Example:
        >>> with open("Prelude.mod", "wb") as file:
        ...    with Writer(file, file_type="MOD ") as e:
        ...        e.add_localized_string(Language.ENGLISH, "Prelude")
        ...        with open("item.uti", "rb") as item:
        ...            e.add_file("item.uti", item)
    """

    class Entry(NamedTuple):
        resref: str
        restype: int
        offset: int
        size: int

    def __init__(
        self,
        file: BinaryIO,
        file_type: FileMagic = FileMagic(b"ERF "),
        build_date=date.today(),
    ):
        self._file = file
        self._entries = []
        self._locstr = {}

        self._file_type = FileMagic(file_type)
        self._build_year = build_date.year - 1900
        self._build_day = build_date.timetuple().tm_yday - 1

    def __enter__(self):
        self._file.write(self._file_type)
        self._file.write(Reader.Version.V_1_0.value.encode("ASCII"))
        self._file.write(b"\x00" * 36)
        self._file.write(b"\x00" * 116)  # reserved bytes as per spec
        assert self._file.tell() == 160
        return self

    def add_localized_string(self, gendered_lang: GenderedLanguage, text):
        self._locstr[gendered_lang] = text

    def add_file(self, filename: str, data: bytes | BinaryIO):
        if isinstance(data, BinaryIO):
            data = data.read()

        offset = self._file.tell()
        size = len(data)
        # ensure we have a restype
        base, ext = filename.split(".")
        if len(base) > 16:
            raise ValueError("Resource name too long")
        rt = extension_to_restype(ext)
        self._entries.append(Writer.Entry(base, rt, offset, size))
        self._file.write(data)
        assert self._file.tell() == offset + size

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            return False

        locstr_offset = self._file.tell()
        for gendered_lang, text in self._locstr.items():
            encoded_text = text.encode(get_nwn_encoding())
            self._file.write(struct.pack("I", gendered_lang.to_id()))
            self._file.write(struct.pack("I", len(encoded_text)))
            self._file.write(encoded_text)
        locstr_size = self._file.tell() - locstr_offset

        keylist_offset = self._file.tell()
        for resref, restype, offset, size in self._entries:
            res_ref = resref.ljust(16, "\x00").encode("ASCII")
            res_id = 0  # res_id unused
            self._file.write(res_ref)
            self._file.write(struct.pack("I", res_id))
            self._file.write(struct.pack("H", restype))
            self._file.write(b"\x00\x00")  # Unused

        reslist_offset = self._file.tell()
        for _, _, offset, size in self._entries:
            self._file.write(struct.pack("i", offset))
            self._file.write(struct.pack("i", size))

        eof_offset = self._file.tell()

        self._file.seek(8)
        self._file.write(
            struct.pack(
                "IIIIIIIII",
                len(self._locstr),
                locstr_size,
                len(self._entries),
                locstr_offset,
                keylist_offset,
                reslist_offset,
                self._build_year,
                self._build_day,
                0,  # description_strref
            )
        )
        self._file.seek(eof_offset)
        return False



from enum import IntEnum
from typing import NamedTuple


class FieldKind(IntEnum):
    BYTE = 0
    CHAR = 1
    WORD = 2
    SHORT = 3
    DWORD = 4
    INT = 5
    DWORD64 = 6
    INT64 = 7
    FLOAT = 8
    DOUBLE = 9
    CEXOSTRING = 10
    RESREF = 11
    CEXOLOCSTRING = 12
    VOID = 13
    STRUCT = 14
    LIST = 15


class Header(NamedTuple):
    file_type: str
    file_version: str
    struct_offset: int
    struct_count: int
    field_offset: int
    field_count: int
    label_offset: int
    label_count: int
    field_data_offset: int
    field_data_size: int
    field_indices_offset: int
    field_indices_size: int
    list_indices_offset: int
    list_indices_size: int


class StructEntry(NamedTuple):
    id: int
    data_or_offset: int
    field_count: int


class FieldEntry(NamedTuple):
    type: FieldKind
    label_index: int
    data_or_offset: int



from io import BufferedReader, BytesIO
from typing import TextIO, BinaryIO
from nwn.nwscript import ndb


class Script:
    """
    Wrapper around NCS/NDB with utility helpers.

    Args:
        ncs_file: The compiled NCS file.
        ndb_file: The NDB debug information.
    """

    def __init__(self, ncs_file: BinaryIO, ndb_file: TextIO):
        self._ncs = ncs_file
        self._ndb = ndb_file

        # We currently cannot easily disasm the NCS file in one go,
        # as the actual bytecode contains relative offset instructions
        # and instructions are of variable length.
        self._ncs_file = BufferedReader(BytesIO(ncs_file.read()[8 + 1 + 4 :]))
        self._ndb_data = ndb.read(ndb_file)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    @classmethod
    def from_compiled(cls, script_name: str):
        """
        Create a Script instance from a filename, relative to cwd.

        Args:
            script_name: The script without file extension.
        """

        return Script(
            open(script_name + ".ncs", "rb"),
            open(script_name + ".ndb", "r"),
        )

    @property
    def ncs(self) -> BufferedReader:
        return self._ncs_file

    @property
    def ndb(self) -> ndb.Ndb:
        return self._ndb_data

    def close(self):
        self._ncs.close()
        self._ndb.close()



"""
Read keyfiles, which store base game resources in the installation directory.
"""

import os
import struct
from datetime import datetime, timedelta, date
from typing import NamedTuple, BinaryIO

from ._shared import restype_to_extension


class _VariableResource(NamedTuple):
    id: int
    io_offset: int
    io_size: int
    res_type: int


class Reader:
    """
    Open a keyfile for reading.

    Example:
        >>> with Reader("nwn_base.key") as rd:
        ...     data = rd.read_file("doortype.2da")
        ...     print(data)

    Args:
        filename: The name of the keyfile to open.
        bif_directory: The directory to search for BIF files in. If not provided,
            the directory containing the keyfile is used.

    Raises:
        ValueError: If the keyfile is not valid.
        FileNotFoundError: If the keyfile or a BIF file is not found.
    """

    class _BIFF(NamedTuple):
        filename: str
        file: BinaryIO
        variable_resources: dict[int, _VariableResource]

    def __init__(self, filename: str, bif_directory=None):
        bif_directory = bif_directory or os.path.dirname(filename) + "/.."
        self._bif_files = {}
        file = self._file = open(filename, "rb")

        magic = file.read(4)
        if magic != b"KEY ":
            raise ValueError("Not a keyfile")
        version = file.read(4)
        if version != b"V1  ":
            raise ValueError("Unsupported keyfile version")

        bif_count = struct.unpack("<I", file.read(4))[0]
        key_count = struct.unpack("<I", file.read(4))[0]
        offset_to_file_table = struct.unpack("<I", file.read(4))[0]
        offset_to_key_table = struct.unpack("<I", file.read(4))[0]
        self._build_year = struct.unpack("<I", file.read(4))[0]
        self._build_day = struct.unpack("<I", file.read(4))[0]

        file.seek(offset_to_file_table)
        file_table = []
        for _ in range(bif_count):
            file_table.append(
                (
                    struct.unpack("<I", file.read(4))[0],
                    struct.unpack("<I", file.read(4))[0],
                    struct.unpack("<H", file.read(2))[0],
                    struct.unpack("<H", file.read(2))[0],
                )
            )

        filename_table = []
        for entry in file_table:
            file.seek(entry[1])
            filename = file.read(entry[2]).decode("ASCII").replace("\\", "/")
            filename_table.append(filename)

        def read_bif(bif_filename):
            bif_file = open(os.path.join(bif_directory, bif_filename), "rb")
            magic = bif_file.read(4)
            if magic != b"BIFF":
                raise ValueError("Not a BIF file")
            version = bif_file.read(4)
            if version != b"V1  ":
                raise ValueError("Unsupported BIF version")
            var_res_count = struct.unpack("<I", bif_file.read(4))[0]
            fixed_res_count = struct.unpack("<I", bif_file.read(4))[0]
            if fixed_res_count != 0:
                raise ValueError("Fixed resources not supported")
            variable_table_offset = struct.unpack("<I", bif_file.read(4))[0]
            variable_resources = {}
            bif_file.seek(variable_table_offset)
            for _ in range(var_res_count):
                full_id = struct.unpack("<I", bif_file.read(4))[0]
                offset = struct.unpack("<I", bif_file.read(4))[0]
                file_size = struct.unpack("<I", bif_file.read(4))[0]
                res_type = struct.unpack("<I", bif_file.read(4))[0]
                variable_resources[full_id & 0xFFFFF] = _VariableResource(
                    id=full_id,
                    io_offset=offset,
                    io_size=file_size,
                    res_type=res_type,
                )
            return self._BIFF(
                os.path.normpath(bif_filename),
                bif_file,
                variable_resources,
            )

        file.seek(offset_to_key_table)
        self._resref_id_lookup = {}
        for _ in range(key_count):
            resref = file.read(16).split(b"\x00")[0].decode("ASCII")
            res_type = struct.unpack("<H", file.read(2))[0]
            res_id = struct.unpack("<I", file.read(4))[0]
            bif_idx = res_id >> 20
            if bif_idx < 0 or bif_idx >= len(file_table):
                raise ValueError("Invalid BIF index")
            resext = restype_to_extension(res_type)
            self._resref_id_lookup[f"{resref}.{resext}"] = res_id

        self._bif_files = [read_bif(fn) for fn in filename_table]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        """
        Closes the main file and all associated BIF files. You should call
        this method when you are done with the keyfile. It is also called
        automatically when the object is deleted (eg. via context manager).
        """

        if self._file:
            self._file.close()
            self._file = None
            for bif_file in self._bif_files:
                bif_file.file.close()
            self._bif_files = []

    @property
    def build_date(self) -> date:
        """
        The build date when this keyfile was created.
        """

        return datetime(1900 + self._build_year, 1, 1).date() + timedelta(
            days=self._build_day - 1
        )

    def filenames(self) -> list[str]:
        """
        Returns all filenames in the keyfile.
        """

        return list(self._resref_id_lookup.keys())

    def read_file(self, filename: str) -> bytes:
        """
        Reads the content of a file from the resource archive.

        Args:
            filename: The name of the file to read, including extension.

        Returns:
            The content of the file.

        Raises:
            ValueError: If the internal state is invalid.
            FileNotFoundError: If the file is not found in the archive.
        """

        res_id = self._resref_id_lookup.get(filename)
        if res_id is None:
            raise FileNotFoundError(f"File {filename} not found in keyfile")
        bif_idx = res_id >> 20
        res_idx = res_id & 0xFFFFF
        bif = self._bif_files[bif_idx]
        resource = bif.variable_resources[res_idx]
        bif.file.seek(resource.io_offset)
        return bif.file.read(resource.io_size)



import struct
from typing import BinaryIO

from nwn._shared import get_nwn_encoding
from nwn.gff._types import (
    Dword64,
    Int64,
    Float,
    Double,
    CExoString,
    ResRef,
    CExoLocString,
    VOID,
    Struct,
    List,
)
from nwn.gff._impl import FieldEntry, StructEntry


def write(file: BinaryIO, root: Struct, file_type: str) -> None:
    """
    Write a GFF data structure to a binary stream.

    The structure must only use supported GFF types, as GFF is a strongly
    typed format and deviation from these will make the game fail to read
    them (e.g. storing a byte as a int will make the game not see it).

    Example:
        >>> root = gff.Struct(0, Byte=gff.Byte(255), SomeStruct=gff.Struct(SomeKey=gff.CExoString("5")))
        ... out = BytesIO()
        ... gff.write(out, root, "TEST")

    Args:
        file: The binary stream to write to.
        root: The root structure of the GFF file.
        file_type: The file type identifier (4 characters).
    """

    labels = []
    fields = []
    structs = []
    list_indices = []
    field_indices = []
    field_data = bytearray()

    label_to_index = {}

    labels = []
    fields = []
    structs = []
    list_indices = []
    field_indices = []
    field_data = bytearray()

    label_to_index = {}

    def _add_label(label: str) -> int:
        binlabel = label.encode("ascii")[0:16]
        binlabel = binlabel.ljust(16, b"\x00")
        if binlabel in label_to_index:
            return label_to_index[binlabel]
        index = len(labels)
        label_to_index[binlabel] = index
        labels.append(binlabel)
        return index

    def _process_field(name: str, value) -> int:
        data_or_offset = len(field_data)

        if hasattr(value, "SIMPLE_DATA_FORMAT"):
            tmp = struct.pack(f"<{value.SIMPLE_DATA_FORMAT}", value)
            padded = tmp.ljust(4, b"\x00")
            data_or_offset = struct.unpack("<I", padded)[0]

        elif isinstance(value, Dword64):
            field_data.extend(struct.pack("<Q", int(value)))

        elif isinstance(value, Int64):
            field_data.extend(struct.pack("<q", int(value)))

        elif isinstance(value, Float):
            data_or_offset = struct.unpack("<I", struct.pack("<f", float(value)))[0]

        elif isinstance(value, Double):
            field_data.extend(struct.pack("<d", float(value)))

        elif isinstance(value, CExoString):
            encoded = value.encode(get_nwn_encoding())
            field_data.extend(struct.pack("<I", len(encoded)))
            field_data.extend(encoded)

        elif isinstance(value, ResRef):
            encoded = value.encode(get_nwn_encoding())
            if len(encoded) > 16:
                raise ValueError("Resref too long")
            field_data.extend(struct.pack("<B", len(encoded)))
            field_data.extend(encoded)

        elif isinstance(value, CExoLocString):
            str_data = bytearray()
            str_data.extend(struct.pack("<I", value.strref))
            str_data.extend(struct.pack("<I", len(value.entries)))
            for fid, text in value.entries.items():
                encoded = text.encode(get_nwn_encoding())
                str_data.extend(struct.pack("<I", fid.to_id()))
                str_data.extend(struct.pack("<I", len(encoded)))
                str_data.extend(encoded)
            field_data.extend(struct.pack("<I", len(str_data)))
            field_data.extend(str_data)

        elif isinstance(value, VOID):
            field_data.extend(struct.pack("<I", len(value)))
            field_data.extend(value)

        elif isinstance(value, List):
            data_or_offset = _process_list(value)

        elif isinstance(value, Struct):
            data_or_offset = _process_struct(value)

        else:
            raise ValueError(f"Field type {type(value)} cannot be serialized to GFF")

        fields.append(
            FieldEntry(
                type=value.__class__.FIELD_KIND,
                label_index=_add_label(name),
                data_or_offset=data_or_offset,
            )
        )

        return len(fields) - 1

    def _process_struct(struct_obj: Struct) -> int:
        structs.append(None)
        struct_index = len(structs) - 1

        struct_field_indices = []
        for name, value in struct_obj.items():
            field_index = _process_field(name, value)
            struct_field_indices.append(field_index)

        if len(struct_field_indices) == 1:
            structs[struct_index] = StructEntry(
                id=struct_obj.struct_id,
                data_or_offset=struct_field_indices[0],
                field_count=1,
            )
        else:
            offset = len(field_indices)
            field_indices.extend(struct_field_indices)
            structs[struct_index] = StructEntry(
                id=struct_obj.struct_id,
                data_or_offset=offset * 4,
                field_count=len(struct_field_indices),
            )

        return struct_index

    def _process_list(list_obj: List) -> int:
        this_list = [len(list_obj)]
        for struct_obj in list_obj:
            struct_index = _process_struct(struct_obj)
            this_list.append(struct_index)
        offset = len(list_indices)
        list_indices.extend(this_list)
        return offset * 4

    root_struct_index = _process_struct(root)
    assert root_struct_index == 0

    header_size = 56
    struct_offset = header_size
    field_offset = struct_offset + len(structs) * 12
    label_offset = field_offset + len(fields) * 12
    field_data_offset = label_offset + len(labels) * 16
    field_indices_offset = field_data_offset + len(field_data)
    list_indices_offset = field_indices_offset + len(field_indices) * 4

    file.write(file_type.encode("ascii"))
    file.write(b"V3.2")

    file.write(
        struct.pack(
            "<12i",
            struct_offset,
            len(structs),
            field_offset,
            len(fields),
            label_offset,
            len(labels),
            field_data_offset,
            len(field_data),
            field_indices_offset,
            len(field_indices) * 4,
            list_indices_offset,
            len(list_indices) * 4,
        )
    )

    for struct_entry in structs:
        file.write(
            struct.pack(
                "<III",
                struct_entry.id,
                struct_entry.data_or_offset,
                struct_entry.field_count,
            )
        )

    for field_entry in fields:
        file.write(
            struct.pack(
                "<III",
                int(field_entry.type),
                field_entry.label_index,
                field_entry.data_or_offset,
            )
        )

    for label in labels:
        file.write(label)

    file.write(field_data)

    for index in field_indices:
        file.write(struct.pack("<I", index))

    for index in list_indices:
        file.write(struct.pack("<I", index))



"""
Read/write nwsync manifests and associated helper utilities.

Binary File Format Version 3
============================

These tables describe the binary file format for the NWSync manifest file itself.

For individual files, see :mod:`nwn.compressedbuf`.

Manifest
--------

===================  ============================================
Field Type           Description
===================  ============================================
uint32               version (always 3)
uint32               count of ManifestEntries
uint32               count of ManifestMappings
ManifestEntry[]      count entries
ManifestMapping[]    count additional mappings
===================  ============================================

ManifestEntry (sorted)
----------------------

===================  ============================================
Field Type           Description
===================  ============================================
byte[20]             SHA1 (as raw bytes) of resref
uint32               size (bytes)
char[16]             resref (WITHOUT extension, pad with 0-bytes)
uint16               restype
===================  ============================================

ManifestMapping (sorted)
------------------------

===================  ============================================
Field Type           Description
===================  ============================================
uint32               Index into ManifestEntry array
char[16]             resref (WITHOUT extension, pad with 0-bytes)
uint16               restype
===================  ============================================
"""

from typing import NamedTuple, BinaryIO

from .._shared import (
    FileMagic,
    restype_to_extension,
    extension_to_restype,
    get_nwn_encoding,
)

NWSYNC_MANIFEST_VERSION = 3
"""The only supported version of the NWSync manifest file format."""

NWSYNC_MANIFEST_MAGIC = FileMagic("NSYM")
"""The file magic for NWSync manifest files."""


class ManifestEntry(NamedTuple):
    sha1: bytes
    """The SHA1 hash of the file, as 20 bytes."""
    size: int
    """The size of the file in bytes."""
    resref: str
    """The full filename, with resolved extension."""

    @property
    def repository_path(self):
        """The relative repository path for the file."""
        sha_hex = self.sha1.hex()
        return f"{sha_hex[0:2]}/{sha_hex[2:4]}/{sha_hex}"


class Manifest(NamedTuple):
    entries: list[ManifestEntry]

    @property
    def version(self):
        """Currently, the only supported version is 3."""
        return NWSYNC_MANIFEST_VERSION


def read(file: BinaryIO) -> Manifest:
    """
    Reads and parses a manifest file from a binary stream.

    Args:
        file: A binary stream representing the manifest file.

    Returns:
        Manifest: An object representing the parsed manifest.

    Raises:
        ValueError: If the file does not contain a valid manifest.
    """

    magic = FileMagic(file.read(4))
    if magic != NWSYNC_MANIFEST_MAGIC:
        raise ValueError("Invalid manifest magic")

    version = int.from_bytes(file.read(4), "little")
    if version != NWSYNC_MANIFEST_VERSION:
        raise ValueError("Unsupported manifest version")

    entry_count = int.from_bytes(file.read(4), "little")
    mapping_count = int.from_bytes(file.read(4), "little")

    if entry_count == 0:
        raise ValueError("Manifest has no entries")

    entries = []
    for _ in range(entry_count):
        sha1 = file.read(20)
        size = int.from_bytes(file.read(4), "little")
        resref = file.read(16).decode(get_nwn_encoding()).rstrip("\0")
        restype = int.from_bytes(file.read(2), "little")
        filename = f"{resref}.{restype_to_extension(restype)}"
        entry = ManifestEntry(sha1, size, filename)
        entries.append(entry)

    for _ in range(mapping_count):
        index = int.from_bytes(file.read(4), "little")
        resref = file.read(16).decode(get_nwn_encoding()).rstrip("\0")
        restype = int.from_bytes(file.read(2), "little")
        filename = f"{resref}.{restype_to_extension(restype)}"
        entry = ManifestEntry(entries[index].sha1, entries[index].size, filename)
        entries.append(entry)

    return Manifest(entries=entries)


def write(file: BinaryIO, manifest: Manifest):
    """
    Writes a manifest to a binary stream.

    Entries are always sorted by SHA1, then by resref, even if the passed-in
    manifest is not sorted. This ensures binary reproducibility.

    Args:
        file: A binary stream to write the manifest to.
        manifest: The manifest to write to the stream.

    Raises:
        ValueError: If the manifest is invalid.
    """

    file.write(NWSYNC_MANIFEST_MAGIC)
    file.write(manifest.version.to_bytes(4, "little"))

    unique_entries = []
    mappings = []
    seen_sha1 = {}

    sorted_entries = [
        (
            entry,
            entry.resref.split(".")[0],
            extension_to_restype(entry.resref.split(".")[1]),
        )
        for entry in manifest.entries
    ]
    sorted_entries.sort(key=lambda x: (x[0].sha1, x[1]))

    for entry, resref_base, restype in sorted_entries:
        if not 0 < len(resref_base) <= 16:
            raise ValueError(f"Resref invalid: {entry.resref}")

        idx = seen_sha1.get(entry.sha1)
        if idx is None:
            seen_sha1[entry.sha1] = len(unique_entries)
            unique_entries.append((entry, resref_base, restype))
        else:
            mappings.append((idx, entry, resref_base, restype))

    file.write(len(unique_entries).to_bytes(4, "little"))
    file.write(len(mappings).to_bytes(4, "little"))

    def write_filename(resref_base, restype):
        file.write(resref_base.encode(get_nwn_encoding()).ljust(16, b"\0"))
        file.write((restype).to_bytes(2, "little"))

    for entry, resref_base, restype in unique_entries:
        file.write(entry.sha1)
        file.write(entry.size.to_bytes(4, "little"))
        write_filename(resref_base, restype)

    for index, entry, resref_base, restype in mappings:
        file.write(index.to_bytes(4, "little"))
        write_filename(resref_base, restype)
